/* $Id: lb.cc,v 1.62 2005/08/02 19:45:32 jonathan Exp $
 * Jonathan Ledlie, Harvard University.
 * Copyright 2005.  All rights reserved.
 */

#include "lb.h"

char *usage = \
"Usage: lb, Distributions\n" \
"            -l physical server lifetime and deathtime file\n" \
"            -H average physical server lifetime\n" \
"            -c physical server capacity file (requests/round/server)\n" \
"            -C average physical server capacity\n" \
"            -w queries per round per alive node\n" \
"            -k routing key distribution\n" \
"\n" \
"           Other parameters\n"
"            -i number of ID choices when joining (default is 1, i.e., off).  Zero is opt\n" \
"            -a [t|p|k] active load balancing method\n" \
"             t: virtual server transfer:\n" \
"               (1) If overloaded and have exactly one virtual server,\n" \
"                   split virtual server into two.\n" \
"               (2) If overloaded and have more than one virtual server,\n" \
"                   attempt to transfer a virtual servers.\n" \
"               Never delete or merge virtual servers.\n" \
"             p: pure virtual servers:\n" \
"               If overloaded this round and have more than one virtual server,\n" \
"               choose the least loaded virtual server that will make us unloaded\n" \
"               and delete it.\n" \
"               If underloaded and adding an average virtual server workload\n" \
"               will not put us over capacity,\n" \
"               create a virtual server\n" \
"             k: continue to use k-choices by actively rechoosing ids\n" \
"              based on perceived workload\n" \
"             g: use Ganesan's Threshold algorithm\n" \
"            -U upper slack pct above which we are overloaded (default .90) \n" \
"            -L lower slack pct below which we are underloaded (default .10)\n" \
"            -s random seed\n" \
"            -e epsilon difference between k-choice keys to make shift\n" \
"            -o oracle mode\n" \
"               - node VSs match object distribution\n" \
"               - all nodes are up at continuously\n" \
"            -d debug\n" \
"            -f frequency of runtime summary output (once every 1000 rounds is default)\n" \
"            -v initial virtual servers per node (1 is default)\n" \
"            -S shift workload to new distribution halfway through run\n" \
"            -O output file prefix\n" \
"            -p k-choices active dampening method\n" \
"               1: limit ids\n" \
"               2: limit create and delete\n" \
"               3: limit both\n" \
"            -m kchoices (active) min capacity, below which rechoosing is dampened\n" \
"\n" \
"Distributions:\n" \
" Pareto       P/scale/shape\n" \
" Normal       N/mean/stddev\n" \
" Zipf         Z/alpha/num-elements\n" \
" Poisson      F/mean\n" \
" Uniform      U [0..1)\n" \
" Constant     C/value\n";

void printUsage () {
  printf ("%s", usage);
  exit (-1);
}

void printUsage (char* problem) {
  printf ("%s\n", problem);
  printUsage ();
}

int main (int argc, char **argv) {
  char *churnFile = NULL, *capacityFile = NULL,
    *keyDistStr = NULL, *outputPrefix = NULL;
  Distribution *keyDist;
  DistributionFactory *distFactory = NULL;
  maxVsPerNode = 128;
  idChoice = 1;
  char activeLBmethod = '-';
  char c = 0;
  int seed = getpid();
  lowerSlack = 0.01;
  upperSlack = 0.99;
#ifdef DEBUG
  debug = false;
#endif
  oracle = false;
  int initialVsPerNode = 1;
  int summaryFrequency = 1000;
  double queriesPerRound = -1.;
  bool workloadShift = false;
  bool finished = false;
  FILE *runLogFP, *linksFP, *summaryFP, *nodeFP;
  int lifetimeAverage = -1, capacityAverage = -1;;
  int dampeningStrategy = 0;

  distFactory = new DistributionFactory ();
  
  while ( (c = getopt(argc,argv,"l:H:C:c:w:k:i:a:s:dU:L:ov:f:SO:e:p:m:")) >= 0) {
    switch (c) {
    case 'l':
      churnFile = optarg;
      break;
    case 'c':
      capacityFile = optarg;
      break;
    case 'w':
      queriesPerRound = atof (optarg);
      break;
    case 'k':
      keyDistStr = optarg;
      break;
    case 'i':
      idChoice = atoi (optarg);
      break;
    case 'p':
      dampeningStrategy = atoi (optarg);
      break;
    case 'H':
      lifetimeAverage = atoi (optarg);
      break;
    case 'C':
      capacityAverage = atoi (optarg);
      break;
    case 'a':
      activeLBmethod = *optarg;
      break;
    case 'o':
      oracle = true;
      break;
    case 's':
      seed = atoi (optarg);
      break;
    case 'U':
      upperSlack = atof (optarg);
      break;
    case 'L':
      lowerSlack = atof (optarg);
      break;
    case 'v':
      initialVsPerNode = atoi (optarg);
      break;
    case 'f':
      summaryFrequency = atoi (optarg);
      break;
    case 'e':
      epsilonImprovement = atof (optarg);
      break;
    case 'm':
      kchoicesActiveMinCapacity = atof (optarg);
      break;
    case 'd':
#ifdef DEBUG
      debug = true;
      break;
#endif
    case 'S':
      workloadShift = true;
      break;
    case 'O':
      outputPrefix = optarg;
      break;
    default:
      printUsage ();
      break;
    }
  }

  SRand (seed);

  switch (dampeningStrategy) {
  case 0:
    break;
  case 1:
    dampeningLimitIds = true;
    break;
  case 2:
    dampeningLimitCrDel = true;
    break;
  case 3:
    dampeningLimitCrDel = true;
    dampeningLimitIds = true;
    break;
  default:
    printUsage ("Problem with dampening strategy");
    break;
  }

  if (lifetimeAverage <= 0)
    printUsage ("Average lifetime must be greater than 0");
  if (capacityAverage <= 0)
    printUsage ("Average capacity must be greater than 0");
  if (upperSlack > 1. || upperSlack < 0)
    printUsage("Upper slack must be between 0 and 1");
  if (lowerSlack > upperSlack || lowerSlack > 1. || lowerSlack < 0)
    printUsage("Lower slack must be between 0 and 1 and less than or equal to upper slack");
  if (activeLBmethod != '-' && activeLBmethod != 't' &&
      activeLBmethod != 'k' && activeLBmethod != 'p' &&
      activeLBmethod != 'g')
    printUsage("Unknown active LB method.");
  if (activeLBmethod == 'g' && initialVsPerNode != 1) {
    printUsage ("Ganesan only allows one initialVsPerNode\n");
  }
  if (idChoice < 0)
    printUsage("Number of IDs must be greater than zero.");
  if (churnFile == NULL)
    printUsage("Error in churn file name");
  if (capacityFile == NULL)
    printUsage("Error in capacity distribution");
  if (keyDistStr == NULL)
    printUsage("Error in key distribution");
  if (initialVsPerNode < 0)
    printUsage("v must be greater or equal to 0 (0 makes log N virtual servers)");
  if (summaryFrequency <= 0)
    printUsage("f must be greater than 0");
  if (queriesPerRound < 0.)
    printUsage("Workload rate must be greater than or equal to 0.");
  if (outputPrefix == NULL)
    printUsage("Output file prefix is required.");

  /*
   * INITIALIZE SYSTEM
   */

  systemTargetCapacity = 0.;
  currentTargetCapacity = 0.;
  recordingStats = false;

  vector<double>capacity;
  PhysicalServer *ps;
  // create PS, set capacities, and create route tables
  psCount = initNodes (capacityFile, ps, initialVsPerNode);

  Statistics *sumStats = new Statistics ();
  Statistics *interimStats = new Statistics ();

  keyDist = distFactory->newDistribution (keyDistStr);
  if (keyDist == NULL)
    printUsage ("Problem with key distribution");

  cRound = 0;

  psUp = 0;
  if (oracle) {
    psCount /= 2;
    if (keyDist->getName() == 'z') {
      oracleAllocateVsZipf (keyDist,ps);
    } else if (keyDist->getName() == 'u') {
      oracleAllocateVsUniform (keyDist,ps);
    } else {
      printUsage("Oracle only handles Zipf and Uniform key distributions.");
    }
    for (int i = 0; i < psCount; i++) {
      ps[i].birth (0);
    }
    psUp = psCount;
  }

  if (activeLBmethod == 'g') {
    usingThreshold = true;
    thresholdSampleSize = (int)(rint(log2(psCount/2)));
  }

  runLogFP = openOutputFile (outputPrefix, ".run");
  traceFP = openOutputFile (outputPrefix, ".trace");
  fprintf (runLogFP,"Earliest LB Round: %d\n", earliestLBround);
  openChurnFile (churnFile);

  while (!finished) {

    DEBUG_P (("ROUND %d\n", cRound));

    /*
     * PER ROUND BIRTH AND DEATH
     */

    // deaths should occur before births to prevent race condition
    char action;
    int nodeId;
    while (!finished && ((action = nextEvent (nodeId)) != ' ')) {
      int vsAct = 0;
      int myInitialVsCount = initialVsPerNode;
      DEBUG_P (("R %d Action: %c node %d\n", cRound, action, nodeId));
      switch (action) {
      case 'b':
	ASSERT(!oracle);
	if (activeLBmethod == 'p') {
	  myInitialVsCount = (int)(ceil (ps[nodeId].getUpperTarget()/10.));
	  if (myInitialVsCount < 1) myInitialVsCount = 1;
	  if (myInitialVsCount > maxVsPerNode) myInitialVsCount = maxVsPerNode;
	}
	vsAct = ps[nodeId].birth (myInitialVsCount);
	interimStats->birth (vsAct);
	break;
      case 'd':
	ASSERT(!oracle);
	vsAct = ps[nodeId].death ();
	interimStats->death (vsAct);
	break;
      case 'r':
	recordingStats = true;
	sumStats->clear();
	printf ("Recording stats.\n");
	break;
      case 'w':
	ASSERT(!oracle);
	printf ("shifting workload\n");
	delete keyDist;
	keyDist = distFactory->newDistribution (keyDistStr);
	if (keyDist == NULL)
	  printUsage ("Problem with key distribution");
	break;
      case 'q':
	finished = true;
	break;
      }
    }
    ASSERT (psUp > 0);

    if (!oracle && cRound == 0) {
      mergeVServers ();
    }

#ifdef DEBUG
    if (debug) {
      printf ("relocated list\n");
      for (set<double>::iterator p = gracefulRelocateVServers.begin();
	   p != gracefulRelocateVServers.end(); p++) {
	printf ("R %d relocated %f\n", cRound, *p);
      }
    }
#endif

      /*
       * PER ROUND VIRTUAL SERVER CREATION, DELETION, SPLITTING, AND TRANSFER
       */

      DEBUG_P (("\nVS ACTIONS: %d\n", cRound));

      for (int i = 0; activeLBmethod != '-' && 
	     !oracle && cRound > earliestLBround && i < psCount; i++) {
	if (ps[i].isAlive()) {
	  if (ps[i].canTakeLBaction()) {
	    int vsCurrentSize = vServers.size();
	    if (activeLBmethod == 't') {
	      if (ps[i].isOverloaded ()) {
		if (ps[i].getVsCount () == 1) {
		  // split can fail if we are adjacent to our neighbor
		  if (ps[i].splitVs ()) {
		    ASSERT (vServers.size() == vsCurrentSize+1);
		    interimStats->splitVs();
		  }
		} else {
		  ASSERT (ps[i].getVsCount () > 1);
		  int vsDelta = 0;
		  if (ps[i].transferVs (vsDelta)) {
		    interimStats->transferVs(true);
		    DEBUG_P (("successful transfer\n"));
		  } else {
		    interimStats->transferVs(false);
		  }
		  if (vServers.size() != vsCurrentSize + vsDelta) {
		    ASSERT (vServers.size() == vsCurrentSize + vsDelta);
		  }
		}
	      }

	    } else if (activeLBmethod == 'p') {
	      if (ps[i].isOverloaded () && ps[i].getVsCount () > 1) {
		int vsDelta = 0;
		ps[i].deleteVs(vsDelta);
		ASSERT (vServers.size() == vsCurrentSize+vsDelta);
		interimStats->deleteVs();
	      } else if (ps[i].isUnderloaded() && ps[i].canAddVs(maxVsPerNode)) {
		DEBUG_P (("PS %d about to createVs\n", i));
		double targetWork = ps[i].getExtraCapacity();
		double predictedWork = 0.;
		if (ps[i].createVs (1,targetWork,predictedWork)) {
		  ASSERT (vServers.size() == vsCurrentSize+1);
		}
		// increment for any kind of activity
		interimStats->createVs();
	      }
	    } else if (activeLBmethod == 'k') {
	      if (ps[i].isOverloaded () || ps[i].isUnderloaded()) {
		int vsDelta = 0;
		int actionCode = ps[i].rechooseVs (vsDelta);
		switch (actionCode) {
		case 0:
		  break;
		case 1:
		  interimStats->createVs();		  
		  break;
		case 2:
		  interimStats->deleteVs();
		  break;
		case 3:
		  interimStats->deleteVs();
		  interimStats->createVs();
		  break;
		default:
		  ASSERT (0);
		}
	      }
	    } else if (activeLBmethod == 'g') {
		int vsDelta = 0;
		int actionCode = ps[i].thresholdVs (vsDelta);
		switch (actionCode) {
		case 0:
		  break;
		case 1:
		  // did neighbor adjust
		  break;
		case 2:
		  // did reorder
		  interimStats->deleteVs();
		  interimStats->createVs();
		  break;
		default:
		  ASSERT (0);
		}
	    }
	  }
	}
      }


    DEBUG_P (("R %d Start of finger updates\n", cRound));
    /*
     * UPDATE FINGER POINTERS
     */
    deadVServersCountVector.push_back ((double)(deadVServers.size()));
    gracefulVServersCountVector.push_back ((double)(gracefulRelocateVServers.size()));
    if (!oracle || (oracle && cRound == 0)) {
      for (map<double,VirtualServer*>::iterator p = vServers.begin();
	   p != vServers.end(); p++) {
        int msgCount = p->second->fixFingers ();
        interimStats->maintMsg (msgCount);
      }
      deadVServers.clear ();
      gracefulRelocateVServers.clear ();
    }
    DEBUG_P (("R %d End of finger updates\n", cRound));

    /*
     * PER ROUND QUERIES
     */

    DEBUG_P (("QUERIES\n"));
    int queryCount = (int)(ceil(psUp * queriesPerRound));
    map<double,VirtualServer*>::iterator p = vServers.lower_bound (randPct());
    vector<double> dstKeys;
    dstKeys.reserve (queryCount);
    for (int i = 0; i < queryCount; i++) {
      double dstKey = keyDist->next();
      dstKeys.push_back (dstKey);
    }
    int queryIndex = -1;
    int querySuccessCount = 0;
    deque<bool> previousQuerySuccess;
    initializePreviousQuerySuccess (previousQuerySuccess);
    for (vector<double>::iterator q = dstKeys.begin(); q != dstKeys.end(); q++) {
      queryIndex++;
      double dstKey = *q;
      ASSERT (dstKey >= 0. && dstKey < 1.);
      DEBUG_P (("Starting query for key %f\n",dstKey));
      // iterate p
      p++;
      if (p == vServers.end()) p = vServers.begin();
      VirtualServer *srcVs = p->second;
      vector<double> hops;
      vector<int> fingersUsed;
      bool unlimitedRouteSuccess = srcVs->findHopDistance (dstKey, hops, fingersUsed, -1.);
      if (0 && hops.size() > 20) {
	printf ("srcVs %f dstVs %f\n", srcVs->getKey(), dstKey);
	int i = 0;
	for (vector<double>::iterator h = hops.begin(); h != hops.end(); h++) {
	  printf ("hop %d %f finger %d\n", i, *h, fingersUsed[i]);
	  i++;
	}
      }
      interimStats->hopSrcDst (unlimitedRouteSuccess, hops.size()-1);
      hops.clear();
      fingersUsed.clear();
      bool routeSuccess = srcVs->route (dstKey, hops, fingersUsed, true);
      int hopCount = hops.size()-1;
      if (hopCount < 0) hopCount = 0;
      if (routeSuccess) {
	DEBUG_P (("R %d query success for key %f in hops %d\n",
		  cRound, dstKey, hopCount));
	interimStats->query (true, hopCount);
        querySuccessCount++;
      } else {
	DEBUG_P (("R %d query fail for key %f\n",cRound,dstKey));
	interimStats->query (false, hopCount);
      }
      interimStats->queryMsg (hopCount);

#ifdef PER_QUERY_STATS
      // per query stats
      if (cRound == 60) {
      double pctRingOK = findPctRingUnderCapacity (ps);
      double hopsEstimate = .5 * log2(psUp);
      double qSEstimate = pow (pctRingOK, hopsEstimate);
      double runningAvgqS = 0;
      double successValue = 0.;
      if (routeSuccess) successValue = 1.;
      runningAvgqS = querySuccessCount/(double)(queryIndex+1);
      addToPreviousQuerySuccess (previousQuerySuccess,routeSuccess);
      double recentTen = avgOfPreviousQuerySuccess (previousQuerySuccess,10);
      double recentTwenty = avgOfPreviousQuerySuccess (previousQuerySuccess,20);

      fprintf (traceFP, "q %d %f real %f est %f ten %f twenty %f pctOK %f\n",
               queryIndex, successValue,
               runningAvgqS,qSEstimate,recentTen,recentTwenty,pctRingOK);
      }
#endif
    }

    IF_DEBUG (printPs (ps,stdout);)

    /*
     * RECORD END OF ROUND
     */
    int workThisRound = 0;
    for (int i = 0; i < psCount; i++) {
      workThisRound += ps[i].endOfRound ();
    }

    interimStats->psUp (psUp);
    interimStats->vsUp (vServers.size());
    cRound++;

    if (cRound % summaryFrequency == 0) {
      if (recordingStats)
	sumStats->add (interimStats);
      fprintf (runLogFP, "R %d ", cRound);
      fprintf (stdout, "R %d ", cRound);
      interimStats->print (runLogFP);
      interimStats->print (stdout);
      interimStats->clear ();
      printUtilHistogram (runLogFP,ps);
      printServerStats (runLogFP,ps);
      printFingerStats (runLogFP);
    }

  } // END OF ROUNDS

  /*
   * PRINT STATS
   */
  // printLinks must come before printPs as it adds up the msgCounts
  sumStats->add (interimStats);
  linksFP = openOutputFile (outputPrefix, ".link");
  summaryFP = openOutputFile (outputPrefix, ".sum");
  nodeFP = openOutputFile (outputPrefix, ".node");

  fprintf (summaryFP, "lF %s lA %d cF %s cA %d q %5.2f k %c %4.3f i %d v %d L %4.3f U %4.3f a %c %d s %d ",
	   churnFile, lifetimeAverage, capacityFile, capacityAverage, queriesPerRound,
	   keyDistStr[0], keyDist->getMean(), idChoice, initialVsPerNode,
	   lowerSlack, upperSlack, activeLBmethod,
	   methodToCode (idChoice, initialVsPerNode, activeLBmethod), seed);
  sumStats->print (summaryFP);
  printLinks (ps, linksFP, false);
  printPs (ps, nodeFP);

  //  printVs (traceFP);

  // destroy PS and route tables

  printf ("Shutting down.\n");
  shuttingDown = true;

  deleteNodes (ps);
  delete keyDist;
  closeChurnFile ();
  delete interimStats;
  delete sumStats;
  delete distFactory;

}



