/* $Id: server.cc,v 1.58 2005/08/02 19:45:32 jonathan Exp $
 * Jonathan Ledlie, Harvard University.
 * Copyright 2005.  All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include "math_util.h"
#include "error.h"
#include "server.h"
#include "lb.h"

double epsilonImprovement = 0.25;
double kchoicesActiveMinCapacity = 10.;

bool debugG = false;
int lookupCount = 0;
int duplicateHop = 0;
int increasingFingerStat = 0;

bool warnedAboutEarliestRound = false;
const int earliestLBround = 400;
const double thresholdDelta = 8.;
const double logOfThresholdDelta = log(thresholdDelta);
const double thresholdInitialConstant = 0.01;
int thresholdSampleSize = 0;
bool usingThreshold = false;

int thresholdAttempt = 0;
int neighborAdjustSuccess = 0;
int reorderAttempt = 0;
int reorderSuccess = 0;

int transferMsgSendFail = 0;
int transferNoCapacityFail = 0;
int transferNoVsMatchFail = 0;
int transferSuccess = 0;
int transferAttempt = 0;

int rechooseAttemptStat = 0;
int rechooseChangeIdStat = 0;
int rechooseCreateStat = 0;
int rechooseDeleteStat = 0;
int rechooseDidNotCreateDeleteStat = 0;
int rechooseOutOfKeysStat = 0;

int rechooseMaxVsSizeStat = 0;
int rechooseWorseCostStat = 0;
bool rechooseFlag = false;

int max_best_keys_size = 0;
int maxHops = 0;
set<double> allKeys;
vector<double> statKeyListSize;


bool shuttingDown;
bool recordingStats;
set<double> deadVServers;
set<double> gracefulRelocateVServers;
map<double,VirtualServer*> vServers;
double systemTargetCapacity;
double currentTargetCapacity;
int lbActionInterval = 30;
int psUp = 0;
int psCount = -1;

int scoreCountStat = 0;
int scorePredictedWorkStat = 0;
int scoreBetterScoreStat = 0;
int scoreBetterDistanceStat = 0;

int routeCountStat = 0;
int routeSendMsgFailedStat = 0;
int routeNextHopIsThisStat = 0;
int routeNullFingerStat = 0;
int routeSendMsgFailedDist[5];
bool initializedRouteSendMsgFailedDist = false;

double getRandomUnusedKey () {
  bool goodId = false;
  double id;
  while (!goodId) {
    goodId = true;
    id = randPct();
    if (allKeys.find(id) != allKeys.end())
      goodId = false;
  }
  return id;
}

double keyDistance (double a, double b) {
  if (a <= b) {
    return b - a;
  }
  return b - a + 1.;
}

double distToSuccessor (VirtualServer *vs) {
  double id = vs->getKey();

  map<double,VirtualServer*>::iterator p;
  p = vServers.lower_bound (id);
  ASSERT (p != vServers.end());
  p++;
  double wrapped = 0.;
  if (p == vServers.end()) {
    p = vServers.begin();
    wrapped = 1.;
  }

  double nextKey = p->second->getKey();

  return nextKey - id + wrapped;
}

/*
 * findDst returns the VS responsible for a key (the clockwise VS).
 * Also used to find a node's successor
 */

VirtualServer* findDst (double key) {
  map<double,VirtualServer*>::iterator p;
  p = vServers.upper_bound (key);
  if (p == vServers.end()) {
    p = vServers.begin();
  }
  return p->second;
}

/*
 * findPred returns the VS of the predesessor of the node at this key
 */
VirtualServer* findPred (double key) {
  map<double,VirtualServer*>::iterator p;
  p = vServers.lower_bound (key);
  if (p == vServers.end()) {
    p = vServers.begin();
  }
  if (p == vServers.begin()) {
    p = vServers.end();
  }
  p--;
  return p->second;
}

/*
 * mergeVServers is called exactly once after the initial node
 * births.  This allows nodes to choose IDs at random for all of 
 * their VSs and then have them merged in one fell swoop.
 */

void mergeVServers () {
  map<double,VirtualServer*>::iterator succ, pred;

  pred = vServers.begin();
  bool reachedEnd = false;
  while (!reachedEnd && pred != vServers.end()) {

    succ = pred;
    succ++;
    if (succ == vServers.end()) {
      reachedEnd = true;
      succ = vServers.begin();
    }

    PhysicalServer *psPred = pred->second->getRootPs();
    PhysicalServer *psSucc = succ->second->getRootPs();

    if (psPred == psSucc) {
      DEBUG_P (("PS %d merging %f into %f\n", 
		psSucc->getId(), pred->first,
		succ->first));
      int vsDelta = 0;
      psPred->deleteLocalVs (pred, vsDelta);
    } else {
      ;
    }
    pred = succ;
  }
}

/*
 * optKey and score are highly related.
 *
 * optKey looks at the gap between each pair of consecutive 
 * VSs and finds the best key/score in that gap.  It returns
 * the key with the best score.
 *
 * score takes a single potential key and the amount of work that
 * the PS would like the VS under formation to perform and returns
 * the score for this particular spot, based on the successor,
 * the size of the gap, etc.
 *
 * optCost computes the minimum cost given target work a, target work b,
 * current work w.  It also returns upper and lower bounds for pct
 * of range that should be given to node a to achieve this cost.
 *
 */

double PhysicalServer::findOptCost (double a, double s,
				    double w, double &min_r, double &max_r) {

  double cost = -1;

  if (s <= 0 && a >= w) {
    printf ("case s<=0 and a>=w\n");
    min_r = 1.;
    max_r = 1.;
    cost = a - s - w;

  } else if (a <= 0 && s >= w) {
    printf ("case s<=0 and a>=w\n");
    min_r = 0.;
    max_r = 0.;
    cost = w - s - a;

  } else if (a + s < w) {
    printf ("case a + s < w\n");
    cost = w - a - s;
    min_r = a/w;
    max_r = 1.-s/w;
  } else {
    //    if (debug)
    printf ("case a + s >= w\n");
    ASSERT (a+s>=w);
    cost = a + s - w;
    min_r = 1.-s/w;
    max_r = a/w;
  }

  cost -= fabs (s-w);

  printf ("a %f s %f w %f cost %f\n", a, s, w, cost);

  return cost;
}

double PhysicalServer::findOptKey (double desired_work, double &best_predicted_work, double &best_cost, double &best_gap) {

  map<double,VirtualServer*>::iterator succ, pred;
  best_cost = 0.;
  bool first_time_flag = true;

  double ourDesiredWork = getExtraCapacity();
  set<OptKeyDesc*> best_keys;

  printf ("PS %d findOptKey\n", id);

  for (succ = vServers.begin(); succ != vServers.end(); succ++) {
    pred = succ;
    if (succ == vServers.begin())
      pred = vServers.end();
    pred--;

    ASSERT (pred->first >= 0. && pred->first < 1.);
    ASSERT (succ->second != NULL);

    printf ("pred %f succ %f succ-id %d succ-cap %f\n", 
	    pred->first, succ->first, succ->second->getRootPs()->getId(),
	    succ->second->rootPs->getCapacity());

    if (succ->second->getRootPs() == this) {
      DEBUG_P (("PS %d skipping ourselves in optKey\n", id));
      continue;
    }

    double key_pred = pred->first;
    double key_succ = succ->first;
    double gap = keyDistance (key_pred,key_succ);

    double succDesiredWork =
      succ->second->getRootPs()->getExtraCapacity() +
      succ->second->getWorkPerRound();

    double succCurrentWork = succ->second->getWorkPerRound();
    double succCapacity = succ->second->rootPs->getCapacity();

    if (succCurrentWork < 0.) {
      succCurrentWork = 0.;
    }

    //desired_work/ourCapacity

    double min_r, max_r;
    double cost = findOptCost (ourDesiredWork,
                               succDesiredWork,
                               succCurrentWork,
                               min_r, max_r);

    printf ("cost %f min_r %f max_r %f\n",
	    cost, min_r, max_r);

    if (first_time_flag || cost <= best_cost) {
      first_time_flag = false;

      //////////////////////////
      // assign a key

      // give a little leeway
      if (max_r > .05) max_r = .95;
      if (min_r < .05) min_r = .05;

      double r = (max_r - min_r) / 2. + min_r;
      ASSERT (r > 0 && r < 1);
      double key = key_pred + r*gap;
      if (key >= 1.) key -= 1.;

      if (vServers.find(key) == vServers.end()) {
        double predicted_work = r * succCurrentWork;
        OptKeyDesc *desc = new OptKeyDesc
          (key, key_succ, key_pred, predicted_work, cost, gap);
        if (cost < best_cost) {
	  printf ("clearing keys\n");
          best_keys.clear ();
        }
	printf ("tentative key %f\n", key);
        best_keys.insert (desc);
      }
    }

    printf ("\n\n");

  }

  if (best_keys.size() > max_best_keys_size) {
    max_best_keys_size = best_keys.size();
    printf ("R %d max_best_keys_size %d\n", cRound, max_best_keys_size);
  }

  int index = unifRand(0,best_keys.size());
  set<OptKeyDesc*>::iterator p = best_keys.begin();
  for (int i = 0; i < index; i++) {
    p++;
  }
  ASSERT (p != best_keys.end());
  OptKeyDesc *opt = *p;
  ASSERT (opt);

  double best_key = opt->key;
  best_predicted_work = opt->predicted_work;
  best_cost = opt->cost;
  best_gap = opt->gap;
  
  ASSERT (best_key >= 0. && best_key < 1.);

  printf ("using key %f\n", best_key);

  return best_key;
}

double PhysicalServer::findDepartureCost (double ourKey, double &gap) {
  map<double,VirtualServer*>::iterator ourIter, succIter, predIter;
  VirtualServer *ourVs, *succVs, *predVs;
  double ourDesiredWork;
  double succDesiredWork, succCurrentWork, succFutureWork;
  double cost;

  ourIter = vServers.find (ourKey);
  ASSERT (ourIter != vServers.end());
  predIter = ourIter;
  if (ourIter == vServers.begin())
    predIter = vServers.end();
  predIter--;

  succIter = ourIter;
  succIter++;  
  if (succIter == vServers.end())
    succIter = vServers.begin();

  ourVs = ourIter->second;
  predVs = predIter->second;
  succVs = succIter->second;
  ourDesiredWork = getExtraCapacity() + ourVs->getWorkPerRound();
  succDesiredWork = succVs->getRootPs()->getExtraCapacity() +
    succVs->getWorkPerRound();
  succCurrentWork = succVs->getWorkPerRound();
  if (succCurrentWork < 0.) succCurrentWork = 0.;
  succFutureWork = ourVs->getWorkPerRound() + succVs->getWorkPerRound();

  DEBUG_P (("R %d PS %d departureCost: us %f (work curr %f des %f) succ %f (work curr %f des %f fut %f)\n",
	    cRound, id, ourVs->getKey(), ourVs->getWorkPerRound(), ourDesiredWork,
	    succVs->getKey(), succCurrentWork, succDesiredWork, succFutureWork));

  // normalized to capacity
  cost = fabs(succFutureWork-succDesiredWork)/succVs->rootPs->getCapacity() -
    fabs(ourVs->getWorkPerRound()-ourDesiredWork)/getCapacity() -
    fabs(succCurrentWork-succDesiredWork)/succVs->rootPs->getCapacity();

  gap = keyDistance (predVs->getKey(), succVs->getKey());
  return cost;
}

double PhysicalServer::findScore (double ourKey, double ourDesiredWork,
			      double &ourFutureWork, double &gap) {
  map<double,VirtualServer*>::iterator succ, pred;
  double ourDomain, succNewDomain, succCurrentDomain;
  double predKey, succKey;
  double succDesiredWork, succCurrentWork, succFutureWork;
  double cost, pctDistance;

  succ = vServers.upper_bound (ourKey);
  pred = succ;
  if (succ == vServers.end()) {
    succ = vServers.begin();
  }
  if (pred == vServers.begin()) {
    pred = vServers.end();
  }
  pred--;

  predKey = pred->first;
  succKey = succ->first;

  succDesiredWork = succ->second->getRootPs()->getExtraCapacity() +
    succ->second->getWorkPerRound();
  succCurrentWork = succ->second->getWorkPerRound();
  if (succCurrentWork < 0.) succCurrentWork = 0.;

  ourDomain = keyDistance (predKey, ourKey);
  succCurrentDomain = keyDistance (predKey, succKey);
  succNewDomain = keyDistance (ourKey, succKey);

  gap = succCurrentDomain;

  pctDistance = ourDomain / succCurrentDomain;

  succFutureWork = (1.-pctDistance)*succCurrentWork;
  ourFutureWork = pctDistance*succCurrentWork;

  cost = fabs(succFutureWork-succDesiredWork)/succ->second->rootPs->getCapacity() +
    fabs(ourFutureWork-ourDesiredWork)/getCapacity() -
    fabs(succCurrentWork-succDesiredWork)/succ->second->rootPs->getCapacity();

#ifdef DEBUG 
  if (debug) {
    printf ("R %d PS %d psUp %d\n", cRound, id, psUp);
    printf ("predKey %f ourKey %f (pctDist %f) succKey %f\n",
	    predKey, ourKey, pctDistance,succKey);
    printf ("succ work: actual: %f desired %f\n",
	    succCurrentWork, succDesiredWork);
    printf ("our work: guess: %f desired %f\n",
	    ourFutureWork, ourDesiredWork);
    printf ("succ work: guess: %f desired %f\n",
	    succFutureWork, succDesiredWork);
    printf ("cost %f\n", cost);

  }
#endif
  ASSERT (pctDistance <= 1. && pctDistance >= 0.);

  return cost;
}

PhysicalServer::PhysicalServer () {

  if (!initializedRouteSendMsgFailedDist) {
    routeSendMsgFailedDist[0] = 0;
    routeSendMsgFailedDist[1] = 0;
    routeSendMsgFailedDist[2] = 0;
    routeSendMsgFailedDist[3] = 0;
    routeSendMsgFailedDist[4] = 0;
    initializedRouteSendMsgFailedDist = true;    
  }

  if (idChoice > 1) {
    for (int i = 0; i < idChoice; i++) {
      double key = getRandomUnusedKey ();
      allKeys.insert (key);
      keys.push_back (key);
    }
  }

  id = 0;
  capacity = 0.;
  upperTarget = 0.;
  lowerTarget = 0.;
  alive = false;
  birthTime = 0;
  deathTime = 0;
  work = 0;
  rechooseAttempts = 0;
  thresholdShift = false;
  thresholdLevel = 0;
  util = 0;
  maxUtil = 0;
  nextLBactionTime = 0;

  nodeStatUptime = 0;
  nodeStatRoundsOverloaded = 0;
  nodeStatTotalWork = 0;
  nodeStatVsAct = 0;
  nodeStatArc = 0;
  nodeStatVsCount = 0;
  nodeStatCreate = 0;
  nodeStatDelete = 0;

  nodeStatRechooseCreate = 0;
  nodeStatRechooseDelete = 0;
  nodeStatRechooseChangeId = 0;

  nodeStatSplit = 0;
  nodeStatTransferAttempt = 0;
  nodeStatTransferSuccess = 0;
  nodeStatQueryIn = 0;
  nodeStatQueryOut = 0;
  nodeStatMaintIn = 0;
  nodeStatMaintOut = 0;

}

PhysicalServer::~PhysicalServer () {
  deleteLocalVServers ();
}

/*
 * PS:setCapacity is called at the beginning of each trial
 * to reset the PS to a known initial state.
 */
void PhysicalServer::setNode (int myId, double c) {
  id = myId;
  alive = false;
  birthTime = -1;
  work = 0;
  util = 0.;
  maxUtil = 0.;
  localVs.clear();
  usedKeys.clear ();
  haveDeletedVS = false;

  nodeStatUptime = 0;
  nodeStatRoundsOverloaded = 0;
  nodeStatTotalWork = 0;

  nodeStatVsAct = 0;
  nodeStatArc = 0.;
  nodeStatVsCount = 0;
  nodeStatCreate = 0;
  nodeStatDelete = 0;
  nodeStatRechooseCreate = 0;
  nodeStatRechooseDelete = 0;
  nodeStatRechooseChangeId = 0;

  nodeStatSplit = 0;
  nodeStatTransferAttempt = 0;
  nodeStatTransferSuccess = 0;
  nodeStatQueryIn = 0;
  nodeStatQueryOut = 0;
  nodeStatMaintIn = 0;
  nodeStatMaintOut = 0;

  if (c <= 0) c = 1;
  this->capacity = c; 
  upperTarget = upperSlack*capacity;
  lowerTarget = lowerSlack*capacity;

  if (upperTarget <= .5) {
    upperTarget = .5;
  }
  if (lowerTarget <= 0.) {
    lowerTarget = 0.;
  }

  if (capacity <= 1.) {
    capacityIndex = 0;
  } else if (capacity <= 10.) {
    capacityIndex = 1;
  } else if (capacity <= 100.) {
    capacityIndex = 2;
  } else if (capacity <= 1000.) {
    capacityIndex = 3;
  } else {
    capacityIndex = 4;
  }
    

}

int util2thresholdLevel (double util) {
  if (util <= thresholdInitialConstant) return 0;
  return (int)(floor(log (util/thresholdInitialConstant)/logOfThresholdDelta));
}

int PhysicalServer::endOfRound () {
  int workThisRound = work;
  if (alive) {
    ASSERT (localVs.size() > 0);
    util = getWorkPerRound() / capacity;

    if (recordingStats) {
      nodeStatUptime++;
      if (util > maxUtil) {
	maxUtil = util;
      }
      if (util > 1.) {
	nodeStatRoundsOverloaded++;
      }
      nodeStatTotalWork += work;

      // can be time consuming
      // nodeStatArc += getArcLength();
      nodeStatVsCount += getVsCount ();
    }
  }

  work = 0;
  for (set<VirtualServer*>::iterator p = localVs.begin();
       p != localVs.end(); p++) {
    (*p)->endOfRound();
  }
  return workThisRound;
}

void VirtualServer::endOfRound () {
  if (workHistory < 0) {
    workHistory = (double)workCurrent;
    DEBUG_P (("R %d PS %d wH neg\n", cRound, rootPs->getId()));
  } else {
    workHistory = EWMA_LAMBDA * workCurrent +
      (1.-EWMA_LAMBDA) * workHistory;

  }  
  DEBUG_P (("R %d PS %d VS %1.10f wH %4.4f wC %5d\n", 
	    cRound, rootPs->getId(), key, workHistory, workCurrent));


  workCurrent = 0;
}

void printUtilHistogram (FILE *fp, PhysicalServer *ps) {
  int b01 = 0, b1 = 0, b10 = 0, b100 = 0, b1000 = 0, bg1000 = 0;
  for (int i = 0; i < psCount; i++) {
    if (ps[i].isAlive()) {
      if (ps[i].getUtil() < 0.1) b01++;
      else if (ps[i].getUtil() < 1.) b1++;
      else if (ps[i].getUtil() < 10) b10++;
      else if (ps[i].getUtil() < 100) b100++;
      else if (ps[i].getUtil() < 1000) b1000++;
      else bg1000++;
    }
  }
  fprintf (fp,"R %d Util: %d [0.1] %d [1] %d [10] %d [100] %d [1000] %d\n",
	  cRound, b01, b1, b10, b100, b1000, bg1000);
}


/*
 * Returns true if our work for this round was above our
 * target upper bound.
 */
bool PhysicalServer::isOverloaded () {
  if (getUtil() > upperSlack) {
    DEBUG_P (("R %d PS %d overloaded\n", cRound, id));
    return true;
  }
  return false;
}

/*
 * Returns true if our work for this round was below our
 * target lower bound.
 */
bool PhysicalServer::isUnderloaded () {
  if (getUtil() < lowerSlack) return true;
  return false;
}

/*
 * Returns the number of virtual servers this PS current has.
 */
int PhysicalServer::getVsCount () {
  return localVs.size();
}

/*
 * Birth of the PS.  Initializes some variables and creates
 * an initial number of VSs.
 */
int PhysicalServer::birth (int initialVsPerNode) {

  int createCount = 0;
  ASSERT (!alive);
  birthTime = cRound;
  deathTime = -1;
  util = 0.;
  usedKeys.clear ();
  haveDeletedVS = false;
  rechooseAttempts = 0;
  thresholdLevel = 0;
  thresholdShift = false;

  setNextLBactionTime ();

  alive = true;
  currentTargetCapacity += (double)(upperTarget);
  systemTargetCapacity += capacity;

  DEBUG_P (("R %d PS %d starting birth\n", cRound, id));

  if (!oracle) {

    double VsDesiredWork = 0.;
    double VsPredictedWork = 0.;

    if (idChoice == 1) {
      for (int i = 0; i < initialVsPerNode; i++) {
	if (createVs (initialVsPerNode, VsDesiredWork, VsPredictedWork)) {
	  createCount++;
	}
      }
    } else {

      int maxToCreate = 3;
      if (maxToCreate > idChoice) maxToCreate = idChoice / 2;
      if (cRound < earliestLBround) maxToCreate = 1;
      bool createOK = true;

      VsDesiredWork = (upperTarget - lowerTarget) / 2. + lowerTarget;
      ASSERT (VsDesiredWork > 0.);

      for (int i = 0; createCount == 0 ||
	     (createOK && VsDesiredWork > 0 &&
	      maxToCreate > i); i++) {
	VsPredictedWork = 0.;

	createOK = createVs (initialVsPerNode, VsDesiredWork, VsPredictedWork);
	if (createOK) {
	  createCount++;
	  VsDesiredWork -= VsPredictedWork;
	}
      }

    }
  }
  psUp++;

  DEBUG_P (("R %d PS %d finished birth arc %f psUp %d cap %f (l %f u %f) lvsSize %d sysCap %f\n",
	    cRound, id, getArcLength(), psUp, capacity,
	    lowerTarget, upperTarget, localVs.size(), currentTargetCapacity));

  if (recordingStats) {
    nodeStatVsCount += createCount;
  }

  // createCount can be 0 if in oracle mode
  return createCount;
}

/*
 * Death of the PS.  All of its VSs are removed from the system.
 */
int PhysicalServer::death () {
  int session = cRound - birthTime;

  DEBUG_P (("PS %d death now %d psUp %d vsSize %d\n",
	    id, cRound, psUp, localVs.size()));
  if (!alive) {
    printf ("PS %d death now %d psUp %d vsSize %d\n",
	    id, cRound, psUp, localVs.size());
    ASSERT (alive);
  }
  alive = false;
  deathTime = cRound;
  psUp--;
  work = 0;

  currentTargetCapacity -= (double)(upperTarget);
  systemTargetCapacity -= capacity;

  // two sanity checks
  if (psUp <= 1) {
    printf ("round is %d\n", cRound);
  }
  ASSERT (psUp > 1);
  ASSERT (currentTargetCapacity > 0.);

  int removedCount = deleteLocalVServers ();
  ASSERT (vServers.size () > 0);

  if (recordingStats) {
    nodeStatVsCount += removedCount;
  }

  return removedCount;
}
 
int PhysicalServer::deleteLocalVServers () {

  int vServerRemoved = localVs.size();
  set<VirtualServer*> localVsCopy (localVs);

  for (set<VirtualServer*>::iterator p = localVs.begin();
       p != localVs.end(); p++) {
    VirtualServer *vs = *p;
    DEBUG_P (("PS %d removing key %f\n", id, vs->getKey()));
    map<double,VirtualServer*>::iterator q = vServers.find ((*p)->getKey());
    int vsDelta = 0;
    removeVs (q, vsDelta);
  }
 
  localVs.clear ();
  return vServerRemoved;
}

/*
 * getRandomKey ()
 *
 * Returns an unused key from this PS's collection
 */
double PhysicalServer::getRandomKey () {
  if (idChoice <= 1) {
    double rid = getRandomUnusedKey ();
    allKeys.insert (rid);
    return rid;
  } else {
    vector<double> unusedKeys = getSampleKeys();
    ASSERT (unusedKeys.size() > 0);
    int randIndex = unifRand (0,unusedKeys.size());
    return unusedKeys[randIndex];
  }
}

/*
 * getSampleKeys ()
 *
 * Returns all unused keys from this PS's collection
 */
vector<double> PhysicalServer::getSampleKeys () {
  vector<double> unusedKeys;
  if (rechooseFlag)
    printf ("sample keys");
  for (int i = 0; i < keys.size(); i++) {
    if (rechooseFlag)
      printf (" %f (%d)", keys[i], i);
    bool valid = true;
    for (set<VirtualServer*>::iterator p = localVs.begin();
	 valid && p != localVs.end(); p++) {
      if ((*p)->getKey() == keys[i]) {
	valid = false;
	if (rechooseFlag)
	  printf ("(in use)");
      }
    }

    if (dampeningLimitIds) {
      for (set<double>::iterator p = usedKeys.begin();
	   valid && p != usedKeys.end(); p++) {
	if (*p == keys[i]) {
	  if (rechooseFlag)
	    printf ("(prior)");
	  valid = false;
	}
      }
    }

    if (valid) {
      unusedKeys.push_back (keys[i]);
    }
  }
  if (rechooseFlag)
    printf ("\n");
  return unusedKeys;
}

/* PS::createVs creates one VS and adds it to 
 * this PS.
 *
 * vsCreateCount is for when more than one Vs is being
 * created at a time (e.g., at birth), and we want each
 * Vs to get one nth of the node's target capacity.
 *
 * The new VSs key is either chosen optimally (idChoice == 0),
 * with the cost function "score", or at random.
 *
 * The function returns false if the PS was the successor to
 * the VS being created and the creation was aborted.
 *
 * If the creation was successful, it returns true.
 *
 * predictedWork is only set if ID selection is turned on.
 */

double PhysicalServer::chooseKey (vector<double> keyList, double desiredWork, 
				  double &predictedWork, double skipKey,
				  double &score, double &gap) {
  double key = -1;
  ASSERT (idChoice != 1);
  ASSERT ((idChoice == 0) || (idChoice > 1 && keyList.size() > 0));

  if (cRound > earliestLBround && idChoice == 0 && vServers.size() > 1) {
    // returns predicted work in desiredWork variable
    key = findOptKey (desiredWork, predictedWork, score, gap);

  } else if (idChoice > 1 &&
	     vServers.size() > 1) {
    if (cRound < earliestLBround) {
      return keyList[0];
    }

    score = -1.;
    predictedWork = -1.;
    gap = 1.;

    DEBUG_P (("PS %d desiredWork %f\n", id, desiredWork));

    for (int i = 0; i < keyList.size(); i++) {
      if (keyList[i] == skipKey) continue;
      double thisPredictedWork = 0.;
      double thisGap = 0.;
      double thisScore = findScore (keyList[i], desiredWork,
                                    thisPredictedWork, thisGap);

      if (predictedWork < 0. || thisScore < score) {

	// all other things being equal, go to the biggest gap

	scoreCountStat++;
	if (predictedWork < 0.) {
	  scorePredictedWorkStat++;
	} else if (thisScore < score) {
	  scoreBetterScoreStat++;
	} else if (thisScore == score &&
		   gap > thisGap) {
	  scoreBetterDistanceStat++;
	}

	key = keyList[i];
	score = thisScore;
	predictedWork = thisPredictedWork;
        gap = thisGap;
        ASSERT (gap > 0. && gap < 1.);
        ASSERT (predictedWork >= 0.);
      }
    }

    DEBUG_P (("PS %d choose key %f target %f future %f score %f sysCap %f\n",
	      id, key, desiredWork, predictedWork,
	      score, systemTargetCapacity));

    ASSERT (predictedWork >= 0.);
  } else {
    key = getRandomKey ();
  }
  return key;
}



bool PhysicalServer::createVs (int vsCreateCount, double desiredWork, 
			       double &predictedWork) {
  ASSERT (alive);
  setNextLBactionTime ();
  double key;
  vector<double> keyList;

  if (idChoice != 1) {
    if (desiredWork <= 0.) {
      // make sure we've created at least one VS for this guy
      ASSERT (localVs.size() > 0);
      return false;
    }

    
    if (idChoice > 1) {
      if (localVs.size() >= idChoice) {
	return false;
      }  
      keyList = getSampleKeys();
    }

    statKeyListSize.push_back ((double)(keyList.size()));
    double cost, gap;
    key = chooseKey (keyList, desiredWork, predictedWork, -1., cost, gap);
  } else {
    key = getRandomKey ();
  }

  if (key < 0. || key >= 1.) {
    printf ("key %f\n", key);
    ASSERT (key >= 0. && key < 1.);
  }

  allKeys.insert (key);

  // This key has already been checked for uniqueness

  bool createdOK = true;

  if (cRound == 0) {
    // Do not merge VSs right at beginning
    DEBUG_P (("PS %d create Vs key %f\n", id, key));
    addVs (key);
  } else {
    map<double,VirtualServer*>::iterator succ, pred;
    succ = vServers.upper_bound (key);
    if (succ == vServers.end()) succ = vServers.begin();

    if (succ->second->getRootPs() != this) {

      // merge, if we own our predecessor
      pred = succ;
      if (pred == vServers.begin()) {
	pred = vServers.end();
      }
      pred--;

      DEBUG_P (("PS %d create Vs key %f\n", id, key));
      VirtualServer* newVs = addVs (key);

      if (pred->second->getRootPs() == this) {
	DEBUG_P (("PS %d merging %f into %f\n", id, pred->first,key));
	newVs->merge (pred->second);
	int vsCount = 0;
	deleteLocalVs (pred, vsCount);
	createdOK = false;
      }
    } else {
      DEBUG_P (("PS %d aborting creation of %f because we are successor\n",
		id, key));
      createdOK = false;
    }
  }

  return createdOK;
}


/*
 * PS::splitVs.
 * Enter with exactly one VS; PS is currently overloaded.
 * Create a new VS at ID that is halfway between current VS
 * and the next VS (that is residing on some other node).
 *
 * If the IDs are two tightly packed and this new "halfway" ID
 * is already in the system, abort the split.  This happens
 * very rarely.
 */

bool PhysicalServer::splitVs () {
  setNextLBactionTime ();
  set<VirtualServer*>::iterator p = localVs.begin();
  VirtualServer *vs = *p;
  ASSERT (localVs.size() == 1 && vs);
  
  double domain = distToSuccessor (vs);
  double key = vs->getKey() + domain / 2.;
  if (key > 1.) key = key - 1.;

  DEBUG_P (("PS %d split Vs newkey %f domain %f\n", id, key, domain));

  // It can happen that there just isn't any room for a key here.
  // In this case, we just return false.
  if (vServers.find(key) != vServers.end()) {
    return false;
  }

  // check that this key isnt currently in use and remove it from
  // free list if it isnt in use
  // it is not possible that the key is in use, because then we would
  // have picked a different key.
  addVs (key);
  if (recordingStats) {
    nodeStatVsCount++;
    nodeStatSplit++;
  }
  return true;
}

/*
 * PS::addVs.  Enter with an allocated, checked key.
 * Do the actual VS creation and adding to local data structures.
 */
VirtualServer* PhysicalServer::addVs (double key) {

  VirtualServer *vs = new VirtualServer (key, this);
  ASSERT (vs);

  vServers.insert(pair<double,VirtualServer*>(key,vs));
  localVs.insert(vs);
  return vs;
}

/*
 * PS::deleteLocalVs is a helper function to remove a VS in merge cases.
 */
void PhysicalServer::deleteLocalVs (map<double,VirtualServer*>::iterator p,
				    int &vsDelta) {
  ASSERT (alive);
  set<VirtualServer*>::iterator q = localVs.find (p->second);
  ASSERT (q != localVs.end());
  removeVs (p, vsDelta);
  localVs.erase (q);
}

/*
 * PS::deleteVs chooses the best VS to delete and deletes it.
 * The best VS is the one most likely to push us below our
 * upper target threshold.
 */
void PhysicalServer::deleteVs (int &vsDelta) {
  ASSERT (alive);
  setNextLBactionTime ();

  // we get here because we are currently overloaded,
  // not because the sum of the VSs histories say we are
  double amtToLose = (double)(upperTarget - getWorkPerRound());

  set<VirtualServer*>::iterator bestVs = localVs.end();
  set<VirtualServer*>::iterator maxVs = localVs.begin();
  for (set<VirtualServer*>::iterator p = localVs.begin();
       p != localVs.end(); p++) {
    VirtualServer *vs = *p;
    if (vs->getWorkPerRound() > amtToLose) {
      if (bestVs == localVs.end() ||
	  (*bestVs)->getWorkPerRound() > vs->getWorkPerRound()) {
	bestVs = p;
      }

    } else if ((*maxVs)->getWorkPerRound() < vs->getWorkPerRound()) {
      maxVs = p;
    }
  }

  if (bestVs == localVs.end()) {
    bestVs = maxVs;
  }
  
  removeVs (vServers.find ((*bestVs)->getKey()), vsDelta);
  localVs.erase (bestVs);      

  // tell those who are pointing to us
  gracefulRelocateVServers.insert ((*bestVs)->getKey());
  return;
}

/*
 * PS::removeLocalVs is used in merge cases where we need to tell
 * another PS that one of its VSs is being removed due to a merge.
 */
void PhysicalServer::removeLocalVs (VirtualServer *vs) {
  localVs.erase (vs);
}

void PhysicalServer::removeVs (map<double,VirtualServer*>::iterator p,
			       int &vsDelta) {
  vsDelta--;
  VirtualServer *vs, *predVs;
  map<double,VirtualServer*>::iterator succ, pred;
  DEBUG_P (("PS %d remove Vs %f\n", id, p->first));
  vs = p->second;
  double key = vs->getKey();
  deadVServers.insert (key);
  succ = p;
  succ++;
  if (succ == vServers.end()) succ = vServers.begin();
  vServers.erase (p);
  delete vs;
  vs = NULL;

  pred = succ;
  if (pred == vServers.begin()) pred = vServers.end();
  pred--;

  if (!shuttingDown && succ->second->getRootPs() == pred->second->getRootPs()) {
    predVs = pred->second;
    DEBUG_P (("PS %d removeVs, PS %d doing a merge: removing %f, merging %f and %f\n", id,
	      predVs->getRootPs()->getId(),
	      key, pred->second->getKey(), succ->second->getKey()));
    //    ASSERT (succ->second->getRootPs()->getId() != id);
    //    set<VirtualServer*>::iterator q = lVs.find (pred->second);
    //    ASSERT (q != lVs.end());
    deadVServers.insert (predVs->getKey());
    vServers.erase (predVs->getKey());
    predVs->getRootPs()->removeLocalVs (predVs);
    delete predVs;
    predVs = NULL;
    vsDelta--;
  }
  return;
}

bool PhysicalServer::sendMaintMsg (double dstKey) {
  VirtualServer *ourSrc = getRandomVs();
  vector<double> hops;
  vector<int> fingersUsed;
  return ourSrc->route (dstKey, hops, fingersUsed, false);
}

/*
 * PS::transferVs picks a node at random and negociates which
 * VS to tranfer to it, if one exists
 */
bool PhysicalServer::transferVs (int &vsDelta) {
  ASSERT (alive);
  if (recordingStats) nodeStatTransferAttempt++;
  transferAttempt++;
  setNextLBactionTime ();

  PhysicalServer *dstPs = this;
  VirtualServer *dstVs = NULL;

  // don't transfer to ourselves
  double dstKey = -1.;
  while ((dstPs == this) && (dstKey = randPct())) {
    dstVs = findDst (dstKey);
    dstPs = dstVs->getRootPs();
  }

  if (!sendMaintMsg (dstKey)) {
    if (debug)
      printf ("R %d PS %d failed to contact random node %f at PS %d to transfer to\n",
              cRound, id, dstKey, dstPs->getId());
    transferMsgSendFail++;
    return false;
  }

  double extraCapacity = dstPs->getExtraCapacity();

  DEBUG_P (("PS %d (work %f, l %d u %d) transfer (randKey %f) to PS %d, extra capacity %f\n",
	    id, getWorkPerRound(), lowerTarget, upperTarget, dstVs->getKey(),
	    dstPs->getId(), extraCapacity));

  if (extraCapacity <= 0.) {
    transferNoCapacityFail++;
    return false;
  }

  set<VirtualServer*>::iterator lightestPtr, heaviestPtr, xferPtr;
  lightestPtr = localVs.end();
  heaviestPtr = localVs.end() ;
  double lightestVsWork = extraCapacity;
  double heaviestVsWork = 0.;

  set<VirtualServer*>::iterator p;
  for (p = localVs.begin(); p != localVs.end(); p++) {
    VirtualServer *vs = *p;
    if (vs->getWorkPerRound() < extraCapacity) {
      if (getWorkPerRound() - vs->getWorkPerRound() < (double)upperTarget) {
	if (vs->getWorkPerRound() < lightestVsWork) {
	  lightestPtr = p;
	}
      } else {
	if (heaviestPtr == localVs.end() ||
	    heaviestVsWork < vs->getWorkPerRound()) {
	  heaviestPtr = p;
	  heaviestVsWork = vs->getWorkPerRound();
	}
      }
    }
  }

  if (lightestPtr != localVs.end()) {
    xferPtr = lightestPtr;
  } else if (heaviestPtr != localVs.end()) {
    xferPtr = heaviestPtr;
  } else {
    transferNoVsMatchFail++;
    return false;
  }
 
  // we are moving the VS along with all of its state
  // along with the VS
  
  (*xferPtr)->setRootPs (dstPs);
  dstPs->takeVs (*xferPtr, vsDelta);
  localVs.erase (xferPtr);

  // incoming pointers need to be notified of new host
  //  gracefulRelocateVServers.insert ((*xferPtr)->getKey());

  if (recordingStats) nodeStatTransferSuccess++;
  transferSuccess++;
  return true;
}

/*
 * PS::takeVs does the work of transferring a VS from one node
 * to another, including merging if necessary.
 */
void PhysicalServer::takeVs (VirtualServer *vs, int &vsDelta) {
  map<double,VirtualServer*>::iterator us, succ, pred;

  ASSERT (vs);

  us = vServers.find (vs->getKey());
  ASSERT (us != vServers.end());
  pred = us;
  if (us == vServers.begin()) {
    pred = vServers.end();
  }
  pred--;
  
  if (pred->second->getRootPs() == this) {
    vs->merge (pred->second);
    deleteLocalVs (pred, vsDelta);
  }

  succ = us;
  succ++;
  if (succ == vServers.end()) {
    succ = vServers.begin();
  }
					  
  if (succ->second->getRootPs() == this) {
    succ->second->merge (vs);
    removeVs (us, vsDelta);
  } else {
    localVs.insert (vs);
  }
  return;
}

VirtualServer* PhysicalServer::getRandomVs () {
  // pick random vs to try
  int index = unifRand(0,localVs.size());

  set<VirtualServer*>::iterator p = localVs.begin();
  for (int i = 0; i < index; i++) {
    p++;
  }
  ASSERT (p != localVs.end());
  VirtualServer *vs = *p;
  ASSERT (vs);
  return vs;
}


// Ganesan's Threshold Algorithm
int PhysicalServer::thresholdVs (int &vsDelta) {
  // hard-coded vsAct:
  // 0 = nothing - threshold failed
  // 1 = did neighbor adjust
  // 2 = did reorder

  thresholdShift = false;
  setNextLBactionTime ();

  ASSERT (localVs.size() == 1);
  set<VirtualServer*>::iterator ourLocalVsIter = localVs.begin();
  VirtualServer *ourVs = *ourLocalVsIter;
  if (ourVs->rootPs->getUtil() < 0) return false;

  thresholdAttempt++;

  // reorder is called from within neighborAdjust
  return neighborAdjust (ourVs);
}

int PhysicalServer::neighborAdjust (VirtualServer *ourVs) {
  // first compare with neighbor
  map<double,VirtualServer*>::iterator ourIter, succIter, predIter,
    moveMeIter;
  bool moveUsTowardPred = false, movePredTowardUs = false;
  ourIter = vServers.find(ourVs->getKey());
  ASSERT (ourIter != vServers.end());

  predIter = ourIter;
  succIter = ourIter;
  succIter++;
  if (succIter == vServers.end())
    succIter = vServers.begin();
  if (predIter == vServers.begin())
    predIter = vServers.end();
  predIter--;
  VirtualServer *predVs = predIter->second;
  VirtualServer *succVs = succIter->second;

  int ourLevel = ourVs->rootPs->getThresholdLevel();
  int succLevel = succVs->rootPs->getThresholdLevel();
  int predLevel = predVs->rootPs->getThresholdLevel();

  if (ourLevel > succLevel && ourLevel > predLevel) {
    if (succLevel > predLevel) {
      // succ has more load than pred
      movePredTowardUs = true;
    } else {
      moveUsTowardPred = true;
    }
  } else if (ourLevel > succLevel) {
    moveUsTowardPred = true;
  } else if (ourLevel > predLevel) {
    movePredTowardUs = true;
  } else {
    if (debugG) 
      printf ("R %d PS %d neighb-adj fail-level p %f %d us %f %d s %f %d\n",
	      cRound, id, predVs->getKey(), predLevel,
	      ourVs->getKey(), ourLevel, succVs->getKey(), succLevel);
    reorderAttempt++;
    if (reorder (ourVs)) {
      return 2;
    }
    return 0;
  }

  double oldKey = -1.;
  double newKey = -1.;
  VirtualServer *moveMeVs = NULL;

  if (moveUsTowardPred) {

    moveMeIter = ourIter;
    moveMeVs = ourVs;
    oldKey = ourVs->getKey();
    newKey = oldKey -
      keyDistance(predVs->getKey(),oldKey)*(1.-1./thresholdDelta);
    if (newKey < 0.) newKey += 1.;
    if (debugG) printf ("moveUsTowardPred: ");

  } else if (movePredTowardUs) {

    moveMeIter = predIter;
    moveMeVs = predVs;
    oldKey = predVs->getKey();
    newKey = oldKey +
      keyDistance(oldKey,ourVs->getKey())*(1.-1./thresholdDelta);
    if (newKey > 1.) newKey -= 1.;
    if (debugG) printf ("movePredTowardUs: ");
  } else {
    ASSERT (0);
  }

  // OK, change the key
  if (vServers.find (newKey) != vServers.end()) {
    // abort if the keys are too tightly packed
    if (debugG) 
      printf ("R %d PS %d neighb-adj fail-key p %1.14f %d us %1.14f %d (tried %1.14f, util %f) s %1.14f %d\n",
	      cRound, id, predVs->getKey(), predLevel, ourVs->getKey(), ourLevel, 
	      newKey, ourVs->rootPs->getUtil(), succVs->getKey(), succLevel);
    return 0;
  }
  vServers.erase (moveMeIter);
  moveMeVs->setKey (newKey);
  vServers.insert(pair<double,VirtualServer*>(newKey,moveMeVs));

  gracefulRelocateVServers.insert (oldKey);
  if (debugG) 
    printf ("R %d PS %d neighb-adj from %1.14f to %1.14f, our key %1.14f, p %d level %d our-level %d s %d level %d\n",
	    cRound, id, oldKey, newKey, ourVs->getKey(),
	    predVs->rootPs->getId(), predLevel, ourLevel, succVs->rootPs->getId(), succLevel);
  neighborAdjustSuccess++;
  return 1;
}

bool PhysicalServer::reorder (VirtualServer *ourVs) {
  // else compare with some random nodes
  
  // The only way for this guy to get a random sample is to 
  // pick numbers at random.
  // He can't pick physical servers uniformly at random.

  double bestUtil = 1.;
  double bestKey = -1.;
  double ourWpr = ourVs->getWorkPerRound();
  VirtualServer *bestVs;
  map<double,VirtualServer*>::iterator bestIter;

  for (int i = 0; i < thresholdSampleSize; i++) {

    double sampleKey = randPct();
    map<double,VirtualServer*>::iterator sampleIter, sampleSuccIter;
    sampleIter = vServers.upper_bound (sampleKey);
    if (sampleIter == vServers.end()) {
      sampleIter = vServers.begin();
    }
    sampleSuccIter = sampleIter;
    sampleSuccIter++;
    if (sampleSuccIter == vServers.end()) {
      sampleSuccIter = vServers.begin();
    }
    VirtualServer *sampleVs = sampleIter->second;
    VirtualServer *sampleSuccVs = sampleSuccIter->second;

    double sampleWpr = sampleVs->getWorkPerRound();
    double sampleSuccWpr = sampleSuccVs->getWorkPerRound();
    if (debugG) 
      printf ("R %d PS %d reorder sample %f w %f l %d succ %f w %f l %d sampleCap %f succCap %f",
	      cRound, id, sampleVs->getKey(), sampleWpr, sampleVs->rootPs->getThresholdLevel(),
	      sampleSuccVs->getKey(), sampleSuccWpr, sampleSuccVs->rootPs->getThresholdLevel(),
	      sampleVs->rootPs->getCapacity(), sampleSuccVs->rootPs->getCapacity());

    // in absolute terms, will this overload the sample's successor?
    if (sampleWpr + sampleSuccWpr > sampleSuccVs->rootPs->getUpperTarget()) {
      if (debugG) printf (" BAD\n");
      continue;
    }

    // in relative terms, pick the one with the lowest util
    double sampleUtil = sampleVs->rootPs->getUtil();
    if (sampleUtil < bestUtil) {
      bestUtil = sampleUtil;
      bestKey = sampleKey;
      bestIter = sampleIter;
      bestVs = sampleVs;
      if (debugG) printf (" BEST\n");
    } else {
      if (debugG) printf (" NOT BEST\n");
    }
  }

  if (bestKey >= 0.) {
    double oldKey = bestKey;
    double ourPredKey = (findPred (ourVs->getKey()))->getKey();
    double newKey = ourPredKey + 
      keyDistance (ourPredKey,ourVs->getKey())/2.;
    if (newKey >= 1.) newKey -= 1.;
    if (vServers.find (newKey) != vServers.end()) {
      if (debugG) 
	printf ("R %d PS %d reorder failed (packed keys) from %f to %f (PS %d)\n",
		cRound, id, oldKey, newKey, bestVs->rootPs->getId());
      return false;
    }
    vServers.erase (bestIter);
    bestVs->setKey (newKey);
    vServers.insert(pair<double,VirtualServer*>(newKey,bestVs));
    gracefulRelocateVServers.insert (oldKey);
    if (debugG) 
      printf ("R %d PS %d reordered from %f to %f (PS %d)\n",
	      cRound, id, oldKey, newKey, bestVs->rootPs->getId());

    reorderSuccess++;
    return true;
  }
  return false;
}

/*
 * When k-choices is active, it calls this function to 
 * reselect IDs and to create or delete VSs
 */
int PhysicalServer::rechooseVs (int &vsDelta) {
  ASSERT (traceFP);
  setNextLBactionTime ();

  bool isOverload = false;
  if (util > upperSlack) {
    isOverload = true;
    if (capacity <= kchoicesActiveMinCapacity && util < 1) return 0;
  } else if (util < lowerSlack) {
    // fuggetabboutit if a low BW node is underloaded
    if (capacity <= kchoicesActiveMinCapacity) return 0;
    ;
  } else {
    // what are we doing here??
    ASSERT (0);
  }

  rechooseAttempts++;
  rechooseAttemptStat++;

  ////////////////////////////////////////////////////////////////////
  // Create or Delete VSs
  if (rechooseAttempts > localVs.size()) {
    rechooseAttempts = 0;
    if (isOverload && localVs.size() > 1) {
      // these are being recorded in lb.cc
      rechooseDeleteStat++;
      if (recordingStats) {
        nodeStatRechooseDelete++;
      }
      haveDeletedVS = true;
      usedKeys.clear();
      deleteVs (vsDelta);
      return 2;
    } else if (!isOverload &&
	       (!dampeningLimitCrDel ||
		!haveDeletedVS)) {
      double targetWork = getExtraCapacity();
      double predictedWork = 0.;
      rechooseCreateStat++;
      if (recordingStats) {
        nodeStatRechooseCreate++;
      }
      usedKeys.clear();
      createVs (1,targetWork,predictedWork);
      return 1;
    } else {
      rechooseDidNotCreateDeleteStat++;
      return 0;
    }
  }

  ////////////////////////////////////////////////////////////////////
  // Move existing VSs to new IDs

  // bestDepCost depicts the change in mismatches
  // without us at that spot anymore.
  // If this is negative, the sum of mismatches has declined

  // we might have no where to move to at all
  if (localVs.size() == idChoice) {
    rechooseMaxVsSizeStat++;
    return 0;
  }

  set<VirtualServer*>::iterator bestVs = localVs.end();

  double bestDepartureCost = -1;
  double departureGap;
  for (set<VirtualServer*>::iterator p = localVs.begin();
       p != localVs.end(); p++) {
    VirtualServer *vs = *p;
    usedKeys.insert (vs->getKey());
    double thisDepartureGap;
    double thisDepartureCost = findDepartureCost (vs->getKey(),thisDepartureGap);
    if (bestVs == localVs.end() ||
	thisDepartureCost < bestDepartureCost) {
      bestVs = p;
      bestDepartureCost = thisDepartureCost;
      departureGap = thisDepartureGap;
    }
  }

  VirtualServer *vs = *bestVs;
  double predictedWork = 0.;
  double desiredWork = getExtraCapacity() + vs->getWorkPerRound();

  DEBUG_P (("R %d PS %d rechooseVs util %f (work %f) localVs %d, key %f departureCost %f desiredWork %f\n",
	    cRound, id, util, getWorkPerRound(), localVs.size(), vs->getKey(), bestDepartureCost, desiredWork));

  double newLocationCost;
  vector<double> unusedKeys = getSampleKeys();
  if (unusedKeys.size() == 0) {
    rechooseOutOfKeysStat++;
    return 0;
  }
  double newLocationGap = 1.;
  double key = chooseKey (unusedKeys, desiredWork,
			  predictedWork, vs->getKey(), newLocationCost, newLocationGap);
  // bestDepartureCost = [future mismatch] - [current mismatch]
  // newLocationCost   = [future mismatch] - [current mismatch]

  if (bestDepartureCost+newLocationCost > -1.* epsilonImprovement) {
    rechooseWorseCostStat++;
    return 0;
  }

  DEBUG_P (("R %d PS %d rechooseVs util %f cur %f new %f curr %f des %f pred %f\n",
	    cRound, id, util, vs->getKey(), key, getWorkPerRound(), desiredWork, predictedWork));

  // it's red five, we're going in

  // relocate in vServers
  map<double,VirtualServer*>::iterator currentLocIter, newLocIter;
  currentLocIter = vServers.find(vs->getKey());
  newLocIter = vServers.find(key);

  ASSERT (currentLocIter != vServers.end());
  if (newLocIter != vServers.end()) {
    printf ("R %d PS %d aborting because oracle picked a key in use\n", cRound, id);
    return 0;
  }
  vServers.erase (currentLocIter);

  // clear fingers and tell it what its new id is
  double oldKey = vs->getKey();
  vs->setKey (key);
  vServers.insert(pair<double,VirtualServer*>(key,vs));

  // notify others
  gracefulRelocateVServers.insert (oldKey);
	  
  // don't need to do anything to localVs
  // not doing merging as chooseKey won't have returned us a key where
  // we would be successor

  rechooseChangeIdStat++;
  if (recordingStats) {
    nodeStatRechooseChangeId++;
  }

  if (idChoice == 0)
    allKeys.insert(key);

  return 3;
}

void PhysicalServer::printLocalVs () {
  for (set<VirtualServer*>::iterator p = localVs.begin();
       p != localVs.end(); p++) {
    VirtualServer *vs = *p;
    fprintf (traceFP, " %f", vs->getKey());
  }
  fprintf (traceFP, "\n");
}

bool PhysicalServer::canTakeLBaction () {
  if (usingThreshold && nextLBactionTime < cRound) {
    if (cRound == birthTime) {
      thresholdLevel = 0;
    } else {
      int currentThresholdLevel = util2thresholdLevel (util);
      if (util > 1. || currentThresholdLevel > thresholdLevel) {
	thresholdShift = true;
      }
      thresholdLevel = currentThresholdLevel;
    }
    ASSERT (thresholdLevel >= 0);
    return thresholdShift;
  }

  if (nextLBactionTime < cRound) return true;
  return false;
}

void PhysicalServer::setNextLBactionTime () {
  if (usingThreshold) nextLBactionTime = cRound + 60;
  nextLBactionTime = cRound + (int)(poisson ((double)lbActionInterval));
}

double VirtualServer::getArcLength () {
#if 0
  map<double,VirtualServer*>::iterator succ =
    vServers.upper_bound (key);
  if (succ == vServers.end()) {
    succ = vServers.begin();
  }
  double succKey = succ->first;
  double arc = keyDistance (key, succKey);
#else
  double arc = keyDistance (findPred(key)->getKey(), key);
#endif
  return arc;
}

double PhysicalServer::getArcLength () {
  if (!alive) return 0.;
  set<VirtualServer*>::iterator p;
  double arc = 0.;
  for (p = localVs.begin(); p != localVs.end(); p++) {
    arc += (*p)->getArcLength();
  }
  return arc;
}

VirtualServer::VirtualServer (double k, PhysicalServer *ps) {
  key = k;
  previousKey = -1.;
  birthTime = cRound;
  workCurrent = 0;
  workHistory = -1.;
  rootPs = ps;
  fingers = new Fingers ();
  predCache = NULL;
}

/*
 * VS::merge merges the histories of two virtual servers into one.
 */
void VirtualServer::merge (VirtualServer *oldGuy) {
  workHistory += oldGuy->getWorkPerRound();
  if (oldGuy->getBirth() < birthTime)
    birthTime = oldGuy->getBirth();
}

VirtualServer::~VirtualServer () {
  delete fingers;
}

void PhysicalServer::incrMaintMsgOut (int mCount) {
  nodeStatMaintOut += mCount;
}

void PhysicalServer::incrMaintMsgIn (int mCount) {
  nodeStatMaintIn += mCount;
}

void PhysicalServer::incrQueryMsgOut (int mCount) {
  nodeStatQueryOut += mCount;
}

void PhysicalServer::incrQueryMsgIn (int mCount) {
  nodeStatQueryIn += mCount;
}


void PhysicalServer::print (FILE *fp) {
  double avgUtil = 0.;
  double avgArc = 0.;
  double avgVsCount = 0.;
  if (nodeStatUptime > 0) {
    avgUtil = (nodeStatTotalWork/(double)nodeStatUptime)/capacity;
    avgArc = nodeStatArc/nodeStatUptime;
    avgVsCount = nodeStatVsCount/(double)nodeStatUptime;
  }
  double avgTransferSuccess = 0.;
  if (nodeStatTransferAttempt > 0) {
    avgTransferSuccess = nodeStatTransferSuccess/(double)avgTransferSuccess;
  }

  set<VirtualServer*>::iterator p = localVs.begin();
  double key = -1.;
  if (p != localVs.end()) {
    VirtualServer *vs = *p;
    key = vs->getKey();
  }

  fprintf (fp, "%-4d cap %5d w %5d up %5d util %8f maxUtil %8f vs %4.1f arc %8f cr %4d del %4d kCr %4d kDel %4d kM %4d spl %4d trA %4d trS %3.2f inQ %5d outQ %5d inM %5d outM %5d key %f\n",
	   id, (int)capacity, nodeStatTotalWork, nodeStatUptime, avgUtil, maxUtil,
	   avgVsCount, avgArc, nodeStatCreate, nodeStatDelete, nodeStatRechooseCreate, nodeStatRechooseDelete, nodeStatRechooseChangeId, nodeStatSplit, nodeStatTransferAttempt,
	   avgTransferSuccess, nodeStatQueryIn, nodeStatQueryOut, nodeStatQueryIn, nodeStatMaintOut, key);
}




/*
 * PS::getExtraCapacity returns the difference between
 * the avg target threshold and the historical average work
 * of this node's VSs.
 */
double PhysicalServer::getExtraCapacity () {
  return (double)(upperTarget+lowerTarget)/2. - getWorkPerRound();
}

/*
 * PS::canAddVs returns true if this node believes it can fit
 * one more VS of average work beneath its upper target threshold.
 */
bool PhysicalServer::canAddVs (int maxVsPerNode) {
  double avgWork = getWorkPerRound();
  if (localVs.size() >= maxVsPerNode) {
    return false;
  }

  if ((double)(upperTarget) > avgWork + avgWork/(double)localVs.size()) {
    return true;
  }
  return false;
}

/*
 * PS::getWorkPerRound finds the historical workload of the combined
 * VSs on this node.
 */
double PhysicalServer::getWorkPerRound () {
  set<VirtualServer*>::iterator p;
  double avgWork = 0.;
  for (p = localVs.begin(); p != localVs.end(); p++) {
    double vsWork = (*p)->getWorkPerRound();
    if (vsWork > 0.) avgWork += vsWork;
  }
  return avgWork;
}

double VirtualServer::getWorkPerRound () {
  return workHistory;
}

/*
 * Returns the percentage of time overloaded over time alive.
 * Note that roundsUp and roundsOverloaded are only incremented
 * when recordingStats is true, which happens the second half of
 * each trial.
 */
/*
  double PhysicalServer::getPctRoundsOverloaded () {
  if (nodeStatUptime == 0) return 0.;
  double pct = (double)roundsOverloaded / (double)roundsUp;
  return pct;
  }
*/

/*
 * PS::getWorkToCapacityRatio returns the historical workload
 * on this node.
 */
double PhysicalServer::getUtil () {
  return util;
}

double PhysicalServer::getMaxUtil () {
  return maxUtil;
}

bool VirtualServer::route (double dstKey, vector<double> &hops, 
			   vector<int> &fingersUsed, bool isQuery) {
  lookupCount++;
  return route (dstKey, hops, fingersUsed, isQuery, -1);
}

bool VirtualServer::route (double dstKey, vector<double> &hops,
                           vector<int> &fingersUsed, 
			   bool isQuery,
			   double dstVsKeyCache) {
  if (hops.size() > maxHops) {
    maxHops = hops.size();
  }

  if (dstVsKeyCache < 0.) {
    VirtualServer *dstVs = findDst (dstKey); 
    dstVsKeyCache = dstVs->getKey();
  }
  DEBUG_P (("PS %d VS %f query dstkey %f\n", rootPs->getId(), key, dstKey));

  if (key == dstVsKeyCache) {
    DEBUG_P (("PS %d VS %f is dst for query of key %f\n",
	      rootPs->getId(), key, dstKey));
    return true;
  }
  DEBUG_P (("PS %d VS %f is not dst vs for key %f, %f is\n",
	    rootPs->getId(), key, dstKey,
	    dstVsKeyCache));
  routeCountStat++;

  int fingerIndex = -1;
  double fingerKey = fingers->findFinger(key,dstKey,fingerIndex,dstVsKeyCache);
  map<double,VirtualServer*>::iterator fingerIter = vServers.find (fingerKey);
  VirtualServer *nextHop = NULL;
  if (fingerIter != vServers.end()) {
    nextHop = fingerIter->second;
  }

  if (!nextHop) {
    DEBUG_P (("PS %d VS %f query finger pointed to dead node\n", 
	      rootPs->getId(), key));
    routeNullFingerStat++;
  }
  if (nextHop == this) {
    // can happen when network is really small
    routeNextHopIsThisStat++;
    printf ("Unusual event: PS %d VS %f not forwarding to ourselves\n", 
	    rootPs->getId(), key);
    return false;
  }

  if (nextHop && nextHop != this) {
    if (nextHop->sendMsg (this,isQuery)) {
      fingers->heartbeat (fingerIndex);
      DEBUG_P (("PS %d VS %f using finger %d to %f hop %d for query of %f\n", 
		rootPs->getId(), key, fingerIndex, nextHop->getKey(), hops.size(), dstKey));
      bool revisit = false;
      int minFinger = 100;
      for (int i = 0; i < hops.size(); i++) {
	if (hops[i] == key) {
	  revisit = true;
	}
	if (fingersUsed[i] < minFinger) {
	  minFinger = fingersUsed[i];
	}
      }

      if (fingerIndex > minFinger) {
	increasingFingerStat++;
	return false;
      }

      if (revisit) {
	printf ("R %d dst %f\n", cRound, dstKey);
	for (int i = 0; i < hops.size(); i++) {
	  printf ("%d hop %f finger %d\n", i, hops[i], fingersUsed[i]);
	}
	printf ("current %f", key);
	printf ("next hop %f", nextHop->getKey());

	duplicateHop++;
	return false;
      }

      hops.push_back(key);
      fingersUsed.push_back (fingerIndex);

      return nextHop->route (dstKey, hops, fingersUsed, isQuery, dstVsKeyCache);
    } else {
      routeSendMsgFailedStat++;
      routeSendMsgFailedDist[nextHop->getRootPs()->getCapacityIndex()]++;
    }
  }

  DEBUG_P (("PS %d VS %f query key %f failed\n", 
	    rootPs->getId(), key, dstKey));
  return false;
}



bool VirtualServer::findHopDistance (double dstKey, vector<double> &hops,
				     vector<int> &fingersUsed, 
				     double dstVsKeyCache) {
  if (dstVsKeyCache < 0.) {
    VirtualServer *dstVs = findDst (dstKey); 
    dstVsKeyCache = dstVs->getKey();
  }
  if (key == dstVsKeyCache) {
    return true;
  }
  int fingerIndex = -1;
  double fingerKey = fingers->findFinger(key,dstKey,fingerIndex,dstVsKeyCache);
  map<double,VirtualServer*>::iterator fingerIter = vServers.find (fingerKey);
  VirtualServer *nextHop = NULL;
  if (fingerIter != vServers.end()) {
    nextHop = fingerIter->second;
  }

  if (nextHop == this) {
    // can happen when network is really small
    return false;
  }

  if (nextHop && nextHop != this) {
    hops.push_back(key);
    fingersUsed.push_back (fingerIndex);
    return nextHop->findHopDistance (dstKey, hops, fingersUsed, dstVsKeyCache);
  }
  return false;
}



bool VirtualServer::sendMsg (VirtualServer *sender, bool isQuery) {
  workCurrent++;
  return rootPs->sendMsg (sender->getRootPs()->getId(), isQuery);
}

bool VirtualServer::sendMsg (int senderId, bool isQuery) {
  workCurrent++;
  return rootPs->sendMsg (senderId, isQuery);
}

bool PhysicalServer::sendMsg (int senderId, bool isQuery) {
  if (!alive) {
    ASSERT (alive);
  }
  ASSERT (senderId >= 0 && senderId < psCount);
  if (recordingStats) {
    if (isQuery)
      queryMsg[senderId][id]++;
    else
      maintMsg[senderId][id]++;
  }
  if (work >= capacity) {
    DEBUG_P (("R %d PS %d got a packet but is overloaded work %d\n",
	      cRound, id, work));
    work++;
    return false;
  }
  DEBUG_P (("R %d PS %d sendMsg work %d\n", cRound, id, work));
  work++;
  return true;
}



bool PhysicalServer::isAlive () {
  return alive;
}

void VirtualServer::setKey (double k) { 
  key = k; 
  workHistory = -1.;
}

OptKeyDesc::OptKeyDesc
(double k, double s, double p, double p_work, double c, double g) {
  key = k;
  succ = s;
  pred = p;
  predicted_work = p_work;
  cost = c;
  gap = g;
}

OptKeyDesc::~OptKeyDesc () {
}


double findPctRingUnderCapacity (PhysicalServer *ps) {

  double psArcs[psCount];
  for (int i = 0; i < psCount; i++) {
    psArcs[i] = 0.;
  }

  for (map<double,VirtualServer*>::iterator walk = vServers.begin();
       walk != vServers.end(); walk++) {
    map<double,VirtualServer*>::iterator nextStep = walk;
    nextStep++;
    if (nextStep == vServers.end())
      nextStep = vServers.begin();
    double distance = keyDistance (walk->first,nextStep->first);
    psArcs[nextStep->second->rootPs->getId()] += distance;
  }

  double okGapPct = 0.;
  for (int i = 0; i < psCount; i++) {
    if (ps[i].isAlive()) {
      if (ps[i].getWork() <= ps[i].getCapacity()) {
        okGapPct += psArcs[i];
      }
    } else {
      ASSERT (psArcs[i] == 0.);
    }
  }
  ASSERT (okGapPct >= 0. && okGapPct <= 1.);
  return okGapPct;
}

void printServerStats (FILE *fp, PhysicalServer *ps) {
  if (!warnedAboutEarliestRound && earliestLBround < 400) {
    printf ("Warning: earliestLBround is %d\n", earliestLBround);
    fprintf (fp, "Warning: earliestLBround is %d\n", earliestLBround);
    warnedAboutEarliestRound = true;
  }

  vector<double> arcs;
  for (map<double,VirtualServer*>::iterator walk = vServers.begin();
       walk != vServers.end(); walk++) {
    map<double,VirtualServer*>::iterator nextStep = walk;
    nextStep++;
    if (nextStep == vServers.end())
      nextStep = vServers.begin();
    double distance = keyDistance (walk->first,nextStep->first);
    arcs.push_back (distance);
  }
  sort(arcs.begin(), arcs.end());  
  fprintf (fp, "R %d VSarcDist: (1) %f (5) %f (10) %f (25) %f (50) %f (75) %f (90) %f (95) %f (99) %f\n",
	   cRound, arcs[(int)(arcs.size()*.01)],
	   arcs[(int)(arcs.size()*.05)], arcs[(int)(arcs.size()*.10)], 
	   arcs[(int)(arcs.size()*.25)], arcs[(int)(arcs.size()*.50)], 
	   arcs[(int)(arcs.size()*.75)], arcs[(int)(arcs.size()*.90)], 
	   arcs[(int)(arcs.size()*.95)], arcs[(int)(arcs.size()*.99)]);


  if (statKeyListSize.size() > 0) {
    double m = mean (statKeyListSize);
    double s = stddev (statKeyListSize,m);
    fprintf (fp,"R %d key list size %f %f\n", cRound, m, s);
  }

  if (routeCountStat > 0) {
    double nullFingerFrac = routeNullFingerStat/(double)routeCountStat;
    //    double nextHopFrac = routeNextHopIsThisStat/(double)routeCountStat;
    double msgFailedFrac = routeSendMsgFailedStat/(double)routeCountStat;
    fprintf (fp,"R %d Queries failing because: nullFinger %f forwardFailed %f\n",
	     cRound, nullFingerFrac, msgFailedFrac);

    if (routeSendMsgFailedStat > 0) {
      double failedOne = routeSendMsgFailedDist[0]/(double)routeSendMsgFailedStat;
      double failedOneToTen = routeSendMsgFailedDist[1]/(double)routeSendMsgFailedStat;
      double failedTenToHundred = routeSendMsgFailedDist[2]/(double)routeSendMsgFailedStat;
      double failedHundredToThousand = routeSendMsgFailedDist[3]/(double)routeSendMsgFailedStat;
      double failedGtThousand = routeSendMsgFailedDist[4]/(double)routeSendMsgFailedStat;

      fprintf (fp, "R %d Queries failing at nodes: %1.5f (1) %1.5f (10) %1.5f (100) %1.5f (1000) %1.5f\n",
	       cRound,
	       failedOne, failedOneToTen, failedTenToHundred,
	       failedHundredToThousand, failedGtThousand);

      routeSendMsgFailedDist[0] = 0;
      routeSendMsgFailedDist[1] = 0;
      routeSendMsgFailedDist[2] = 0;
      routeSendMsgFailedDist[3] = 0;
      routeSendMsgFailedDist[4] = 0;
    }

    routeNullFingerStat = 0;
    routeCountStat = 0;
    routeNextHopIsThisStat = 0;
    routeSendMsgFailedStat = 0;

  }

  int capOne = 0;
  int capOneToTen = 0;
  int capTenToHundred = 0;
  int capHundredToThousand = 0;
  int capGtThousand = 0;
  int capTotal = 0;
  for (int i = 0; i < psCount; i++) {
    if (ps[i].isAlive()) {
      capTotal++;
      switch (ps[i].getCapacityIndex()) {
      case 0:
	capOne++;
	break;
      case 1:
	capOneToTen++;
	break;
      case 2:
	capTenToHundred++;
	break;
      case 3:
	capHundredToThousand++;
	break;
      case 4:
	capGtThousand++;
	break;
      default:
	ASSERT (0);
      }
    }
  }

  double capOneFrac = capOne/(double)capTotal;
  double capOneToTenFrac = capOneToTen/(double)capTotal;
  double capTenToHundredFrac = capTenToHundred/(double)capTotal;
  double capHundredToThousandFrac = capHundredToThousand/(double)capTotal;
  double capGtThousandFrac = capGtThousand/(double)capTotal;

  fprintf (fp, "R %d cap-dist: %d [%f] (1) %d [%f] (10) %d [%f] (100) %d [%f] (1000) %d [%f] \n",
	   cRound, capOne, capOneFrac, capOneToTen, capOneToTenFrac, capTenToHundred, capTenToHundredFrac,
	   capHundredToThousand, capHundredToThousandFrac, capGtThousand, capGtThousandFrac);


  if (rechooseAttemptStat > 0) {
    double changeIdPct = rechooseChangeIdStat/(double)rechooseAttemptStat;
    double createPct = rechooseCreateStat/(double)rechooseAttemptStat;
    double deletePct = rechooseDeleteStat/(double)rechooseAttemptStat;
    double outOfKeysPct = rechooseOutOfKeysStat/(double)rechooseAttemptStat;
    double neitherPct = rechooseDidNotCreateDeleteStat/(double)rechooseAttemptStat;
    double maxVsPct = rechooseMaxVsSizeStat/(double)rechooseAttemptStat;
    double worseCostPct = rechooseWorseCostStat/(double)rechooseAttemptStat;
    fprintf (fp,"R %d Times in rechoose: %d Pct we changed IDs %f Pct we add %f del %f no add/del %f maxVs %f out-of-keys %f worseCost %f\n",
	    cRound, rechooseAttemptStat, changeIdPct, createPct, deletePct, neitherPct,
	    maxVsPct, outOfKeysPct, worseCostPct);
    rechooseAttemptStat = 0;
    rechooseChangeIdStat = 0;
    rechooseCreateStat = 0;
    rechooseDeleteStat = 0;
    rechooseDidNotCreateDeleteStat = 0;
    rechooseMaxVsSizeStat = 0;
    rechooseWorseCostStat = 0;
  }

  if (thresholdAttempt > 0) {
    double reorderSuccessRate = 0.;
    if (reorderAttempt > 0.) reorderSuccessRate = reorderSuccess/(double)reorderAttempt;
    double neighborAdjustSuccessRate = neighborAdjustSuccess/(double)thresholdAttempt;
    fprintf (fp,"R %d Threshold attemptCount %d neighAdjPct %f reorderPct %f \n",
	    cRound, thresholdAttempt,
	    neighborAdjustSuccessRate, 
	    reorderSuccessRate);
    reorderAttempt = 0;
    reorderSuccess = 0;
    thresholdAttempt = 0;
  }

  if (transferAttempt > 0) {
    double msgFailPct = transferMsgSendFail/(double)transferAttempt;
    double noCapacityFailPct = transferNoCapacityFail/(double)transferAttempt;
    double noVsMatchFailPct = transferNoVsMatchFail/(double)transferAttempt;
    double successPct = transferSuccess/(double)transferAttempt;
    fprintf (fp, "R %d Transfer attemptCount %d no-cap %f no-match %f succ %f\n",
             cRound, transferAttempt, noCapacityFailPct, noVsMatchFailPct, successPct);
    transferMsgSendFail = 0;
    transferNoCapacityFail = 0;
    transferNoVsMatchFail = 0;
    transferSuccess = 0;
    transferAttempt = 0;
  }

  if (scoreCountStat > 0) {
    double predWorkFrac = scorePredictedWorkStat/(double)scoreCountStat;
    double betterScoreFrac = scoreBetterScoreStat/(double)scoreCountStat;
    double betterDistanceFrac = scoreBetterDistanceStat/(double)scoreCountStat;
    fprintf (fp,"R %d Score: predwork %f betterscore %f betterdistance %f\n",
	    cRound, predWorkFrac, betterScoreFrac, betterDistanceFrac);
    scoreCountStat = 0;
    scorePredictedWorkStat = 0;
    scoreBetterScoreStat = 0;
    scoreBetterDistanceStat = 0;
  }

  if (lookupCount > 0) {
    double dupPct = duplicateHop/(double)lookupCount;
    double incrPct = increasingFingerStat/(double)lookupCount;
    fprintf (fp,"R %d Lookup: maxHops %d increasingFinger %f (bad) dupPct %f\n", cRound, maxHops, incrPct, dupPct);
    duplicateHop = 0;
    increasingFingerStat = 0;
  }

  fflush (fp);
  if (traceFP != NULL) {
    fflush (traceFP);
  }

}

